local addonName, a = ...
local L = a.Localize
local s = SpellFlashAddon
local x = s.UpdatedVariables
local c = BittensGlobalTables.GetTable("BittensSpellFlashLibrary")
local m = a.CatSimulator

local GetComboPoints = GetComboPoints
local GetEclipseDirection = GetEclipseDirection
local GetMeleeHaste = GetMeleeHaste
local GetPowerRegen = GetPowerRegen
local GetTime = GetTime
local IsMounted = IsMounted
local SPELL_POWER_ENERGY = SPELL_POWER_ENERGY
local SPELL_POWER_ECLIPSE = SPELL_POWER_ECLIPSE
local SPELL_POWER_RAGE = SPELL_POWER_RAGE
local math = math
local select = select
local string = string
local print = print
local tostring = tostring

local RIP_ACCEPTABLE_DOWNTIME = 2
local RIP_FURY_DELAY = 2
local RAKE_FURY_DELAY = 2

a.Rotations = {}

----------------------------------------------------------------------- Balance
a.Rotations.Balance = {
	Spec = 1,
	
	FlashInCombat = function()
		a.Energy = s.Power("player", SPELL_POWER_ECLIPSE)
		if c.HasBuff("Celestial Alignment") then
			a.Solar = true
			a.Lunar = true
		else
			a.Solar = c.HasBuff("Eclipse (Solar)") 
			a.Lunar = c.HasBuff("Eclipse (Lunar)")
		end
		a.GoingUp = GetEclipseDirection() ~= "moon"
		
		a.EclipsePending = false
		local info = c.GetCastingInfo()
		if info and not c.HasBuff("Celestial Alignment") then
			local bump = 0
			if c.InfoMatches(info, "Starfire") then
				if a.GoingUp then
					bump = 20
				end
			elseif c.InfoMatches(info, "Wrath") then
				if not a.GoingUp then
					bump = -15
				end
			elseif c.InfoMatches(info, "Starsurge") then
				if a.GoingUp then
					bump = 20
				else
					bump = -20
				end
			end
			if s.HasSpell(c.GetID("Euphoria"))
				and not a.Solar
				and not a.Lunar then
				
				bump = bump * 2
			end
			a.Energy = a.Energy + bump
			if a.Energy <= -100 then
				a.GoingUp = true
				a.Lunar = true
				a.EclipsePending = true
			elseif a.Energy >= 100 then
				a.GoingUp = false
				a.Solar = true
				a.EclipsePending = true
			elseif not c.HasBuff("Celestial Alignment") then
				if a.Energy >= 0 then
					a.Lunar = false
				end
				if a.Energy <= 0 then
					a.Solar = false
				end
			end
		end
		
		c.FlashAll(
			"Moonkin Form", 
			"Starfall", 
			"Force of Nature", 
			"Incarnation: Chosen of Elune",
			"Nature's Vigil",
			"Celestial Alignment",
			"Renewal",
			"Solar Beam")
		
		c.PriorityFlash(
			"Starsurge under Shooting Stars",
			"Moonfire under Eclipse",
			"Sunfire under Eclipse",
			"Moonfire",
			"Sunfire",
			"Starsurge",
			"Starfire under Celestial Alignment",
			"Wrath under Celestial Alignment",
			"Starfire",
			"Wrath")
	end,
	
	FlashOutOfCombat = function()
		c.FlashAll("Symbiosis", "Healing Touch Solo")
	end,
	
	FlashAlways = function()
		c.FlashAll("Mark of the Wild")
	end,
	
	ExtraDebugInfo = function()
		return string.format("%d, %s, %s, %s, %s",
			a.Energy,
			tostring(a.Solar),
			tostring(a.Lunar),
			tostring(a.GoingUp),
			tostring(a.EclipsePending))
	end,
}

------------------------------------------------------------------------- Feral
local lastRipTime = 0
local lastRipDuration = 0
local pendingRipExtension = 0

local function updateRipDuration()
	local duration = s.MyDebuffDuration(c.GetID("Rip"))
	local now = GetTime()
	if duration - lastRipDuration > 4 then
		pendingRipExtension = 6
--c.Debug("Rip", "Overwrite: ", duration - lastRipDuration)
	elseif pendingRipExtension > 0 and duration > lastRipDuration then
		pendingRipExtension = pendingRipExtension - 2
--c.Debug("Rip", "change of ", duration - lastRipDuration, " -> ", pendingRipExtension)
	end
	lastRipTime = now
	lastRipDuration = duration
	
	local busyTime = c.GetBusyTime()
	if c.IsAuraPendingFor("Rip") then
		a.Rip = 99
	elseif duration > busyTime 
		or (duration > 0 
			and pendingRipExtension > 0 
			and c.IsCasting("Shred", "Ravage", "Mangle(Cat Form)")) then
		
		a.Rip = duration - busyTime
		if not c.AoE then
			a.Rip = a.Rip + pendingRipExtension
		end
	else
		a.Rip = 0
	end
end

local lastEnergy = 0
local pendingBiteDrain = 0
local forestPending = 0
local biteTime = 0
local cpPending = 0
local furyPending = 0
local pendingRoarDrop = 0
local lastCP = 0

local function initState()
	if s.Power("player") < lastEnergy or GetTime() - biteTime > 1 then
		pendingBiteDrain = 0
	end
	lastEnergy = s.Power("player")
	
	updateRipDuration()
	
	local now = GetTime()
	local curCP = GetComboPoints("player")
	local added = curCP - lastCP
	a.CP = curCP
	if now - cpPending < .8 then
		if added > 0 then
			added = added - 1
			cpPending = 0
			c.Debug("Event", "Mangle CP Happened")
		else
			a.CP = a.CP + 1
		end
	end
	if now - furyPending < .8 then
		if added > 0 then
			furyPending = 0
			c.Debug("Event", "Primal Fury Happened")
		else
			a.CP = a.CP + 1
		end
	end
	if now - pendingRoarDrop < .8 then
		if curCP == 0 then
			pendingRoarDrop = 0
			c.Debug("Event", "Savage Roar CP Use Happened")
		else
			a.CP = 0
		end
	end
	a.CP = math.min(5, a.CP)
	lastCP = curCP
	
	a.Rage = c.GetPower(0, SPELL_POWER_RAGE)
	
	a.Regen = 10 + GetMeleeHaste() / 10
	a.Energy = 
		c.GetPower(a.Regen, SPELL_POWER_ENERGY) 
		- pendingBiteDrain 
		+ forestPending
	a.Roar = c.GetBuffDuration("Savage Roar", false, false, true)
	a.Rake = c.GetMyDebuffDuration("Rake", false, false, true)
	if c.IsCasting("Healing Touch") then
		a.Swiftness = false
		if c.HasTalent("Dream of Cenarius") then
			a.DreamStacks = 2
		end
	else
		a.Swiftness = c.HasBuff("Predatory Swiftness") 
			or c.HasBuff("Nature's Swiftness", false, false, true) 
		a.DreamStacks = c.GetBuffStack(
			"Dream of Cenarius Damage", false, true, "Healing Touch")
		if a.DreamStacks > 0 
			and c.IsCasting(
				"Shred", 
				"Shred!",
				"Thrash(Bear Form)",
				"Mangle(Bear Form)",
				"Rake", 
				"Ravage", 
				"Ravage!",
				"Rip", 
				"Ferocious Bite", 
				"Thrash(Cat Form)", 
				"Mangle(Cat Form)",
				"Swipe") then
			
			a.DreamStacks = a.DreamStacks - 1
		end
		if c.IsCasting(
			"Shred", 
			"Shred!", 
			"Rake", 
			"Ravage", 
			"Ravage!", 
			"Mangle(Cat Form)") then
			
			a.CP = math.min(5, a.CP + 1)
		elseif c.IsCasting("Rip", "Savage Roar", "Ferocious Bite") then
			if a.CP == 5 then
				a.Swiftness = true
			end
			if c.HasTalent("Soul of the Forest") then
				a.Energy = a.Energy + 4 * a.CP
			end
			
			a.CP = 0
		end
	end
	a.Energy = math.max(0, math.min(100, a.Energy))
	
	a.Clearcasting = c.HasBuff("Clearcasting")
	if a.Clearcasting 
		and c.IsCasting(
			"Shred",
			"Shred!",
			"Rake",
			"Ravage",
			"Ravage!",
			"Rip",
			"Ferocious Bite",
			"Thrash(Cat Form)",
			"Mangle(Cat Form)",
			"Swipe") then
		
		a.Clearcasting = false
	end
	
	a.Berserk = c.GetBuffDuration("Berserk")
	if c.IsCasting("Berserk") then
		a.Berserk = 99
	end
	
	a.TigersFury = c.GetBuffDuration("Tiger's Fury", false, false, true)
	a.King = c.GetBuffDuration(
		"Incarnation: King of the Jungle", false, false, true)
end

a.Rotations.Feral = {
	Spec = 2,
	
	FlashInCombat = function()
		initState()
		
		c.FlashAll("Survival Instincts under 30", "Barkskin under 30")
		
		if s.Form(c.GetID("Bear Form")) then
			c.FlashAll("Heart of the Wild")
			local flashing = c.PriorityFlash(
				"Swipe(Bear Form) Prime for Feral",
				"Thrash(Bear Form) for Feral",
				"Cat Form",
				"Swipe(Bear Form) for Feral",
				"Mangle(Bear Form)",
				"Lacerate",
				"Faerie Fire")
			if a.Rage >= (c.AoE and flashing ~= "Cat Form" and 45 or 30) then
				c.FlashAll("Maul for Feral") 
			end
			return
		end
		
		if c.FlashAll("Cat Form") then
			return
		end
		
		c.FlashAll(
			"Faerie Fire for Debuff",
			"Tiger's Fury",
			"Berserk",
			"Incarnation: King of the Jungle",
			"Force of Nature",
			"Nature's Vigil",
			"Healing Touch for Feral Heal",
			"Renewal",
			"Skull Bash")
		if c.AoE then
			c.PriorityFlash(
				"Healing Touch for Dream",
				"Savage Roar for AoE",
				"Rip",
				"Ferocious Bite for AoE",
				"Nature's Swiftness pre-Thrash",
				"Bear Form for Feral AoE",
				"Thrash(Cat Form) for AoE",
				"Swipe(Cat Form) for Feral")
		else
			c.PriorityFlash(
				"Healing Touch for Dream",
				"Savage Roar",
				"Ferocious Bite on Last Tick",
				"Nature's Swiftness pre-Rip",
				"Rip",
				"Savage Roar Early",
				"Nature's Swiftness pre-Bite",
				"Ferocious Bite",
				"Rake",
				"Thrash(Cat Form)",
				"Healing Touch for Vigil",
				"Nature's Swiftness for Vigil",
				"Rip Delay",
				"Bear Form for Feral",
				"Ravage",
				"Shred",
				"Mangle(Cat Form)")
		end
	end,
	
	FlashOutOfCombat = function()
		c.FlashAll("Symbiosis", "Healing Touch Solo")
	end,
	
	FlashAlways = function()
		c.FlashAll("Mark of the Wild")
	end,
	
	CastQueued = function(info)
		if c.InfoMatches(info, "Ferocious Bite") then
			local _, regen = GetPowerRegen()
			local cost = s.SpellCost(info.Name)
			local energy = s.Power("player") + s.UpdatedVariables.Lag * regen
			info.Cost = cost + math.min(25, energy - cost)
			c.Debug("Event", "FB will cost", info.Cost)
--		elseif c.InfoMatches(info, "Healing Touch") and a.Swiftness then
--			info.
		end
	end,
	
	CastSucceeded = function(info)
		local hasForest = c.HasTalent("Soul of the Forest")
		
		if c.InfoMatches(info, "Mangle(Cat Form)") then
			cpPending = GetTime()
			c.Debug("Event", "Mangle CP Pending")
		elseif c.InfoMatches(info, "Savage Roar Glyphed") then
			pendingRoarDrop = GetTime()
			c.Debug("Event", "Savage Roar CP Use Pending")
		elseif c.InfoMatches(info, "Ferocious Bite") then
			lastEnergy = s.Power("player")
			if hasForest then
				pendingBiteDrain = math.min(
					25, lastEnergy - 4 * GetComboPoints("player"))
			else
				pendingBiteDrain = math.min(25, lastEnergy)
			end
			biteTime = GetTime()
			c.Debug("Event", "FB will still cost", pendingBiteDrain)
		end
		
		if hasForest
			and c.InfoMatches(
				info, 
				"Rip",
				"Savage Roar Glyphed", 
				"Savage Roar Unglyphed") then
			
			forestPending = 4 * GetComboPoints("player")
			c.Debug("Event", "Soul of the Forest Pending:", forestPending)
		end
	end,
	
	SpellMissed = function(spellID)
		if c.IdMatches(spellID, "Rip", "Ferocious Bite") then
			forestPending = 0
			c.Debug("Event", "Never mind on Soul of the Forest ... it missed")
		end
		if c.IdMatches(spellID, "Ferocious Bite") then
			pendingBiteDrain = 0
			c.Debug("Event", "Oh, and never mind on FB either")
		end
	end,
	
	Energized = function(spellID)
		if c.IdMatches(spellID, "Soul of the Forest") then
			forestPending = 0
			c.Debug("Event", "Soul of the Forest Happened")
		end
	end,
	
	SpellDamage = function(spellID, _, _, crit, tick)
		if crit
			and not tick
			and c.IdMatches(
				spellID, 
				"Shred", 
				"Shred!", 
				"Rake",
				"Ravage",
				"Ravage!",
				"Mangle(Cat Form)") then
			
			furyPending = GetTime()
			c.Debug("Event", "Primal Fury Pending")
		end
	end,
	
	ExtraDebugInfo = function()
		return string.format(
			"r:%.1f e:%.1f c:%d c:%s s:%s d:%d b:%.1f t:%.1f k:%.1f r:%.1f r:%.1f b:%.1f", 
			a.Rage,
			a.Energy, 
			a.CP,
			tostring(not not a.Clearcasting),
			tostring(not not a.Swiftness),
			a.DreamStacks,
			a.Berserk,
			a.TigersFury,
			a.King,
			a.Rip,
			a.Roar,
			c.GetBusyTime())
	end,
}

---------------------------------------------------------------------- Guardian
local uncontrolledMitigationCooldowns = {
	"Feint",
	"Spell Reflection",
	"Tooth and Claw",
}

a.Rotations.Guardian = {
	Spec = 3,
	
	FlashInCombat = function()
		a.Rage = c.GetPower(select(2, GetPowerRegen()))
		if c.IsQueued("Mangle(Bear Form)") then
			local bump = 5
			if c.HasTalent("Soul of the Forest") then
				bump = bump + 3
			end
			if c.HasBuff("Enrage", true) and c.WearingSet(4, "GuardianT15") then
				bump = bump * 1.5
			end
			a.Rage =  math.min(s.MaxPower("player"), a.Rage + bump)
		end
		a.EmptyRage = s.MaxPower("player") - a.Rage
		
		c.FlashMitigationBuffs(
			1,
			uncontrolledMitigationCooldowns,
			"Cenarion Ward for Guardian",
			"Bone Shield",
			"Elusive Brew",
			"Barkskin",
			"Nature's Swiftness for Guardian",
			"Flashing Steel Talisman",
			"Renewal for Guardian",
			"Survival Instincts Glyphed",
			"Might of Ursoc at 2 mins",
			"Berserk for Guardian",
			"Incarnation: Son of Ursoc",
			"Survival Instincts Unglyphed",
			"Might of Ursoc above 2 mins")
		c.FlashAll(
			"Bear Form",
			"Frenzied Regeneration",
			"Savage Defense",
			"Maul for Guardian",
			"Skull Bash",
			"Growl")
		c.DelayPriorityFlash(
			"Thrash(Bear Form) for Weakened Blows",
			"Healing Touch for Guardian",
			"Rejuvenation for Guardian",
			"Mangle(Bear Form) for Guardian",
			"Mangle(Bear Form) Delay",
			"Rejuvenation Refresh for Guardian",
			"Swipe(Bear Form) for AoE",
			"Swipe(Bear Form) for AoE Delay",
			"Thrash(Bear Form) for AoE",
			"Thrash(Bear Form) for Bleed",
			"Faerie Fire for Debuff for Guardian",
			"Enrage",
			"Lacerate for Guardian",
			"Faerie Fire",
			"Thrash(Bear Form)")
	end,
	
	FlashOutOfCombat = function()
		c.FlashAll("Symbiosis", "Healing Touch Solo", "Bone Shield")
	end,
	
	FlashAlways = function()
		c.FlashAll("Mark of the Wild")
	end,
	
	ExtraDebugInfo = function()
		return string.format("r:%.1f b:%.2f", a.Rage, c.GetBusyTime())
	end,
}

------------------------------------------------------------------- Restoration
a.Rotations.Restoration = {
	Spec = 4,
	
	FlashInCombat = function()
		c.FlashAll(
			"Lifebloom",
			"Swiftmend",
			"Nourish",
			"Healing Touch for Restoration",
			"Regrowth",
			"Innervate",
			"Renewal",
			"Cenarion Ward for Restoration")
	end,
	
	FlashAlways = function()
		c.FlashAll("Mark of the Wild")
	end,
	
	AuraApplied = function(spellID, target)
		if c.IdMatches(spellID, "Lifebloom") then
			a.LifebloomTarget = target
			c.Debug("Event", "Lifebloom on", target)
		end
	end,
}
