local addonName, a = ...
local L = a.Localize
local s = SpellFlashAddon
local c = BittensGlobalTables.GetTable("BittensSpellFlashLibrary")

local GetComboPoints = GetComboPoints
local UnitAttackPower = UnitAttackPower
local UnitInRange = UnitInRange
local UnitIsUnit = UnitIsUnit
local UnitStat = UnitStat
local select = select
local math = math

------------------------------------------------------------------------ Common
local function substantialFight()
	return s.Health() > 1.5 * s.MaxHealth("player")
end

local function setFaerieID(z)
	z.Name = nil
	if c.HasTalent("Faerie Swarm") then
		z.ID = c.GetID("Faerie Swarm")
	else
		z.ID = c.GetID("Faerie Fire")
	end
end

c.AddOptionalSpell("Mark of the Wild", nil, {
	Override = function()
		return c.RaidBuffNeeded(c.STAT_BUFFS)
	end
})

c.AddOptionalSpell("Symbiosis", nil, {
	Override = function()
		if c.GetBuffDuration(a.SymbiosisSelfBuffs) > 5 * 60 then
			return false
		end
		for member in c.GetGroupMembers() do
			if not UnitIsUnit(member, "player") 
				and UnitInRange(member) 
				and (s.MyBuff(a.SymbiosisRaidBuffs, member)
					or not s.Buff(a.SymbiosisRaidBuffs, member)) then
				
				return true
			end
		end
	end
})

c.AddSpell("Faerie Fire", nil, {
	RunFirst = setFaerieID,
	Cooldown = 6,
})

c.AddOptionalSpell("Faerie Fire", "for Debuff", {
	Debuff = c.ARMOR_DEBUFFS,
	EarlyRefresh = 5,
	RunFirst = setFaerieID,
	CheckFirst = substantialFight,
})

c.AddOptionalSpell("Healing Touch", "Solo", {
	NoRangeCheck = true,
	CheckFirst = function()
		return c.GetHealthPercent("player") < 75
	end
})

c.AddOptionalSpell("Force of Nature", nil, {
	NoRangeCheck = true,
})

c.AddOptionalSpell("Nature's Vigil", nil, {
	NoRangeCheck = true,
})

c.AddOptionalSpell("Renewal", nil, {
	NoGCD = true,
	CheckFirst = function()
		return c.GetHealthPercent("player") < 70
	end,
})

c.AddInterrupt("Skull Bash", nil, {
	NoGCD = true,
})

----------------------------------------------------------------------- Balance
local function notCastThisEclipse(name)
	return a.EclipsePending
		or c.GetMyDebuffDuration(name) < c.GetBuffDuration("Nature's Grace") - 2
end

c.AddOptionalSpell("Moonkin Form", nil, {
	Type = "form",
})

c.AddOptionalSpell("Incarnation: Chosen of Elune", nil, {
	CheckFirst = function()
		return a.Energy >= 100 or a.Energy <= -100
	end
})

c.AddOptionalSpell("Celestial Alignment", nil, {
	CheckFirst = function()
		return not (a.Lunar or a.Solar)
			and c.GetCooldown("Incarnation: Chosen of Elune") > 10
	end
})

c.AddOptionalSpell("Starfall", nil, {
	Override = function(z)
		if c.GetCooldown("Incarnation: Chosen of Elune") > 15 then
			z.FlashSize = nil
		elseif not a.GoingUp
			and (a.Energy <= -70 
				or (c.GetCooldown("Starsurge") == 0 and a.Energy <= -60)) then
			z.FlashSize = s.FlashSizePercent() / 2
		else
			return false
		end
		return not c.HasBuff("Starfall")
			and (c.GetCooldown("Starfall") == 0
				or (a.EclipsePending and a.Lunar))
	end
})

c.AddSpell("Starsurge", nil, {
	NotIfActive = true,
})

c.AddSpell("Starsurge", "under Shooting Stars", {
	CheckFirst = function()
		return c.HasBuff("Shooting Stars")
	end
})

c.AddSpell("Moonfire", nil, {
	CheckFirst = function()
		return notCastThisEclipse("Moonfire")
	end
})

c.AddSpell("Moonfire", "under Eclipse", {
	EarlyRefresh = 99,
	CheckFirst = function(z)
		return a.Lunar 
			and (notCastThisEclipse("Moonfire") 
				or c.GetMyDebuffDuration("Moonfire") < z.EarlyRefresh)
	end
})
c.ManageDotRefresh("Moonfire under Eclipse", 1, "Moonfire")

c.AddSpell("Sunfire", nil, {
	CheckFirst = function()
		return notCastThisEclipse("Sunfire")
	end
})

c.AddSpell("Sunfire", "under Eclipse", {
	EarlyRefresh = 99,
	CheckFirst = function(z)
		return a.Solar 
			and (notCastThisEclipse("Sunfire")
				or c.GetMyDebuffDuration("Sunfire") < z.EarlyRefresh)
	end
})
c.ManageDotRefresh("Sunfire under Eclipse", 1, "Sunfire")

c.AddSpell("Wrath", "under Celestial Alignment", {
	CheckFirst = function()
		return c.GetBuffDuration("Celestial Alignment") 	
			> c.GetCastTime("Wrath") + .1
	end
})

c.AddSpell("Starfire", nil, {
	CheckFirst = function()
		return a.GoingUp
	end
})

c.AddSpell("Starfire", "under Celestial Alignment", {
	CheckFirst = function()
		return c.GetBuffDuration("Celestial Alignment") 	
			> c.GetCastTime("Starfire") + .1
	end
})

c.AddInterrupt("Solar Beam")

------------------------------------------------------------------------- Feral
local function getDelay(cost, usesCC)
	if usesCC and a.Clearcasting then
		return 0
	end
	
	if a.Berserk > 0 then
		local delay = (cost / 2 - a.Energy) / a.Regen
		if delay < a.Berserk then
			return math.max(0, delay)
		end
	end
	
	return math.max(0, (cost - a.Energy) / a.Regen)
end

local function modForDelay(spell, cost, usesCC)
	local delay = getDelay(cost, usesCC)
	c.MakePredictor(spell, delay > 0)
	return delay
end

c.AddSpell("Survival Instincts", "under 30", {
	NoGCD = true,
	FlashColor = "red",
	CheckFirst = function()
		return c.GetHealthPercent("player") < 30
	end
})

c.AddSpell("Barkskin", "under 30", {
	NoGCD = true,
	FlashColor = "red",
	CheckFirst = function()
		return c.GetHealthPercent("player") < 30
	end
})

c.AddOptionalSpell("Maul", "for Feral", {
	Melee = true,
	NoGCD = true,
	Override = function(z)
		c.MakePredictor(z, c.GetCooldown("Maul", true) > 0)
		return true
	end,
})

c.AddOptionalSpell("Heart of the Wild", nil, {
	Melee = true,
	FlashSize = s.FlashSizePercent() / 2,
})

c.AddSpell("Thrash(Bear Form)", "for Feral", {
	FlashID = { "Thrash", "Thrash(Bear Form)", "Thrash(Cat Form)" },
	Melee = true,
	NotWhileActive = true,
	CheckFirst = function()
		return (c.GetMyDebuffDuration("Thrash(Bear Form)", false, true, true)
					< 2
				and substantialFight())
			or c.AoE
	end,
})

c.AddSpell("Thrash(Bear Form)", "for Feral AoE", {
	FlashID = { "Thrash", "Thrash(Bear Form)", "Thrash(Cat Form)" },
	Melee = true,
	Cooldown = 6,
	CheckFirst = substantialFight,
})

c.AddSpell("Mangle(Bear Form)", "for Feral", {
	FlashID = { "Mangle", "Mangle(Bear Form)", "Mangle(Cat Form)" },
	CheckFirst = function()
		return (100 - a.Energy) / a.Regen > (a.Clearcasting and 4 or 3)
	end
})

c.AddSpell("Cat Form", nil, {
	Type = "form",
	CheckFirst = function(z)
		-- also adjust conditions in Swipe(Bear Form) Prime
		if not s.Form(c.GetID("Bear Form")) 
			or (100 - a.Energy) / a.Regen < (a.Clearcasting and 4 or 3) then
			
			return true
		elseif c.AoE then
			return c.GetMyDebuffDuration("Thrash(Cat Form)", false, true) < 3
		else
			return a.Rip < (a.CP == 0 and 4 or 3) and substantialFight()
		end
	end,
})

c.AddOptionalSpell("Tiger's Fury", nil, {
	NoGCD = true,
	CheckFirst = function()
		return a.Energy <= 35 
			and not a.Clearcasting
			and not c.HasBuff("Berserk", true)
	end
})

c.AddOptionalSpell("Berserk", nil, {
	NoGCD = true,
	CheckFirst = function(z)
		local furyCool = c.GetCooldown("Tiger's Fury")
		if furyCool > 6 or not substantialFight() then
			z.FlashColor = "yellow"
		else
			z.FlashColor = "green"
		end
		
		if furyCool == 0 and substantialFight() then
			z.FlashSize = s.FlashSizePercent() / 2
		else
			z.FlashSize = nil
		end
		
		return furyCool > 6 or furyCool == 0 or not substantialFight()
	end
})

c.AddOptionalSpell("Incarnation: King of the Jungle")

c.AddSpell("Healing Touch", "for Dream", {
	Override = function()
		return a.DreamStacks <= 0
			and a.Swiftness
			and a.Energy + a.Regen < 100
			and c.HasTalent("Dream of Cenarius")
	end
})

c.AddSpell("Healing Touch", "for Vigil", {
	Override = function()
		return a.Swiftness
			and a.Energy < 35
			and c.HasBuff("Nature's Vigil", false, false, true)
	end
})

c.AddOptionalSpell("Healing Touch", "for Feral Heal", {
	Override = function()
		return c.GetHealthPercent("player") < 85
			and a.Swiftness 
			and not c.HasTalent("Dream of Cenarius")
	end
})

local function setRoarId(spell)
	if c.HasGlyph("Savagery") then
		spell.ID = c.GetID("Savage Roar Glyphed")
	else
		spell.ID = c.GetID("Savage Roar Unglyphed")
	end
end

c.AddSpell("Savage Roar", nil, {
	RunFirst = setRoarId,
	Override = function(z)
		return (a.CP > 0 or c.HasGlyph("Savagery"))
			and modForDelay(z, 25) >= a.Roar
	end,
})

c.AddSpell("Savage Roar", "Early", {
	RunFirst = setRoarId,
	Override = function(z)
		return a.CP == 5
			and a.Roar < a.Rip + 2
			and a.Roar - modForDelay(z, 25) < 6
			and substantialFight()
	end
})

c.AddSpell("Savage Roar", "for AoE", {
	RunFirst = setRoarId,
	Override = function(z)
		return (a.CP > 0 or c.HasGlyph("Savagery"))
			and (modForDelay(z, 25) >= a.Roar
				or (a.CP == 5 and a.Roar < 10))
	end,
})

c.AddSpell("Ferocious Bite", nil, {
	Melee = true,
	Override = function(z)
		local delay = modForDelay(z, 25, true)
		return a.CP == 5
			and (a.InExecute or a.Rip - delay > 6 or not substantialFight())
	end
})

c.AddSpell("Ferocious Bite", "on Last Tick", {
	Melee = true,
	Override = function(z)
		local rip = a.Rip - modForDelay(z, 25, true)
		return a.InExecute and a.CP > 0 and rip > 0 and rip < 2
	end
})

c.AddSpell("Ferocious Bite", "for AoE", {
	Melee = true,
	Override = function(z)
		local delay = modForDelay(z, 25, true)
		return a.CP == 5
			and a.InExecute
			and a.Rip - delay < 10 
			and substantialFight()
	end
})

local function swiftable()
	return not a.Swiftness
		and a.DreamStacks == 0
		and a.CP == 5
		and a.Energy + a.Regen < 100
		and c.HasTalent("Dream of Cenarius")
		and substantialFight()
end

local function flashSwift(condition)
	if condition and c.GetCooldown("Nature's Swiftness") == 0 then
		s.Flash(c.GetID("Healing Touch"), "green", s.FlashSizePercent() / 2)
		return true
	end
end

c.AddOptionalSpell("Nature's Swiftness", "pre-Rip", {
	Override = function()
		return flashSwift(
			swiftable() and (not a.InExecute or a.Rip == 0) and a.Rip < 3.5)
	end
})

c.AddOptionalSpell("Nature's Swiftness", "pre-Bite", {
	Override = function()
		return flashSwift(swiftable() and a.InExecute and a.Rip < 9.5)
	end
})

c.AddOptionalSpell("Nature's Swiftness", "pre-Thrash", {
	Override = function()
		return flashSwift(swiftable() 
			and c.GetMyDebuffDuration("Thrash(Cat Form)", false, true)
					- getDelay(50, true)
				< 3)
	end
})

c.AddOptionalSpell("Nature's Swiftness", "for Vigil", {
	Override = function()
		return flashSwift(
			a.Energy < 35
				and c.HasBuff("Nature's Vigil", false, false, true)
				and c.GetCooldown("Tiger's Fury") > 0)
	end
})

c.AddSpell("Rip", nil, {
	Melee = true,
	Override = function(z)
		return a.CP == 5
			and (not a.InExecute or a.Rip == 0)
			and a.Rip - modForDelay(z, 30, true) < 2
			and substantialFight()
	end
})

c.MakePredictor(c.AddSpell("Rip", "Delay", {
	Melee = true,
	Override = function()
		return not a.InExecute
			and a.CP == 5
			and (100 - a.Energy) / a.Regen > a.Rip + 1
			and substantialFight()
	end
}))

c.AddSpell("Rake", nil, {
	Melee = true,
	Override = function(z)
		local dur = a.Rake - modForDelay(z, 35, true)
		return dur < 3 or (dur < 6 and a.DreamStacks == 1 and a.CP >= 4)
	end
})

c.AddSpell("Thrash(Cat Form)", nil, {
	FlashID = { "Thrash", "Thrash(Bear Form)", "Thrash(Cat Form)" },
	Melee = true,
	Override = function()
		return getDelay(50, true) == 0
			and c.GetMyDebuffDuration("Thrash(Cat Form)", false, true, true) < 3
			and (a.Clearcasting or a.Rip > 7 or a.CP == 5)
			and substantialFight()
	end
})

c.AddSpell("Thrash(Cat Form)", "for AoE", {
	FlashID = { "Thrash", "Thrash(Bear Form)", "Thrash(Cat Form)" },
	Melee = true,
	Override = function(z)
		return c.GetMyDebuffDuration("Thrash(Cat Form)", false, true, true)
				- modForDelay(z, 50, true) 
			< 3
	end
})

c.AddSpell("Ravage", nil, {
	FlashID = { "Ravage", "Ravage!" },
	Melee = true,
	Override = function(z)
		return modForDelay(z, 45, true) < a.King
	end
})

c.AddSpell("Shred", nil, {
	FlashID = { "Shred", "Shred!" },
	Melee = true,
	Override = function(z)
		local delay = modForDelay(z, 40, true)
		local _, targetID = s.UnitInfo()
		return a.Regen > 15
			and (not (a.NoShred[targetID] or c.IsTanking())
				or (c.HasGlyph("Shred") 
					and (a.TigersFury > delay or a.Berserk > delay)))
	end
})

c.AddSpell("Mangle(Cat Form)", nil, {
	FlashID = { "Mangle", "Mangle(Bear Form)", "Mangle(Cat Form)" },
	Melee = true,
	Override = function(z)
		modForDelay(z, 35, true)
		return true
	end
})

c.AddOptionalSpell("Bear Form", "for Feral", {
	Type = "form",
	Melee = true,
	NotWhileActive = true,
	CheckFirst = function()
		if a.Energy >= 35
			or (100 - a.Energy) / a.Regen < 4
			or a.Berserk > 0 then
			
			return false
		elseif c.AoE then
			return c.GetMyDebuffDuration("Thrash(Cat Form)", false, false, true)
					> 4.5
				and c.GetCooldown("Tiger's Fury") > 4.5
		elseif substantialFight() then
			return a.Rip > (a.CP == 0 and 4.5 or 5.5)
					and a.Rake > 4.5
					and c.GetCooldown("Tiger's Fury") > 4.5
		else
			return c.GetCooldown("Tiger's Fury") > 0
		end
	end
})

c.AddOptionalSpell("Bear Form", "for Feral AoE", {
	Type = "form",
	Melee = true,
	NotWhileActive = true,
	CheckFirst = function()
		return a.Energy < 45
			and (100 - a.Energy) / a.Regen > 4
			and a.Berserk == 0 
			and c.GetMyDebuffDuration("Thrash(Cat Form)", false, false, true)
					> 4
			and c.GetCooldown("Tiger's Fury") > 4
	end
})

c.AddSpell("Swipe(Bear Form)", "for Feral", {
	FlashID = { "Swipe", "Swipe(Bear Form)", "Swipe(Cat Form)" },
	Melee = true,
	NotWhileActive = true,
	CheckFirst = function()
		return c.AoE
	end,
})

c.AddSpell("Swipe(Bear Form)", "Prime for Feral", {
	FlashID = { "Swipe", "Swipe(Bear Form)", "Swipe(Cat Form)" },
	Melee = true,
	NotWhileActive = true,
	CheckFirst = function()
		return c.AoE
			and (100 - a.Energy) / a.Regen > (a.Clearcasting and 5.5 or 4.5) 
			and c.GetMyDebuffDuration("Thrash(Cat Form)", false, true) > 4.5
	end,
})

c.AddSpell("Swipe(Cat Form)", "for Feral", {
	FlashID = { "Swipe", "Swipe(Bear Form)", "Swipe(Cat Form)" },
	Melee = true,
	Override = function(z)
		modForDelay(z, 45, true)
		return true
	end
})

---------------------------------------------------------------------- Guardian
local function damageMode()
	return c.IsSolo() and c.GetHealthPercent("player") > 50
end

c.AddOptionalSpell("Bear Form", nil, {
	Type = "form",
})

c.AddOptionalSpell("Tooth and Claw", nil, {
	NoGCD = true,
	IsUp = function()
		return c.HasMyDebuff("Tooth and Claw", true)
	end,
})

c.AddOptionalSpell("Cenarion Ward", "for Guardian", {
	NoRangeCheck = true,
	CheckFirst = function()
		return not damageMode()
	end,
})

c.AddOptionalSpell("Bone Shield", nil, {
	FlashID = a.SymbiosisGuardianFlashIDs,
	NoRangeCheck = true,
	CheckFirst = function()
		local stack = c.GetBuffStack("Bone Shield")
		if s.InCombat() then
			return stack == 0
		else
			return stack < 3 or c.GetBuffDuration("Bone Shield") < 20
		end
	end,
})

c.AddOptionalSpell("Elusive Brew", nil, {
	FlashID = a.SymbiosisGuardianFlashIDs,
	NoRangeCheck = true,
	NoGCD = true,
})

c.AddOptionalSpell("Barkskin", nil, {
	NoGCD = true,
})

c.AddOptionalSpell("Nature's Swiftness", "for Guardian", {
	NoRangeCheck = true,
	InactiveFlashID = { "Healing Touch", "Nature's Swiftness" },
	RunFirst = function(z)
		if c.HasBuff("Nature's Swiftness") then
			z.ID = c.GetID("Healing Touch")
			z.FlashID = nil
		else
			z.ID = c.GetID("Nature's Swiftness")
			z.FlashID = z.InactiveFlashID
		end
	end, 
	ShouldHold = function()
		return c.GetHealthPercent("player") > 80
	end,
})

c.AddOptionalSpell("Flashing Steel Talisman", nil, {
	Type = "item",
	NoRangeCheck = true,
	Buff = "Flashing Steel",
	BuffUnit = "player",
})

c.AddOptionalSpell("Renewal", "for Guardian", {
	NoGCD = true,
	ShouldHold = function()
		return c.GetHealthPercent("player") > 70
	end,
})

c.AddOptionalSpell("Survival Instincts", "Unglyphed", {
	NoGCD = true,
	Enabled = function()
		return not c.HasGlyph("Survival Instincts")
	end,
	CheckFirst = function()
		return a.Rage > 90 or not c.HasBuff("Frenzied Regeneration", true)
	end,
})

c.AddOptionalSpell("Survival Instincts", "Glyphed", {
	NoGCD = true,
	Enabled = function()
		return c.HasGlyph("Survival Instincts")
	end,
	CheckFirst = function()
		return a.Rage > 90 or not c.HasBuff("Frenzied Regeneration", true)
	end,
})

c.AddOptionalSpell("Might of Ursoc", "at 2 mins", {
	NoGCD = true,
	IsUp = function()
		return c.HasBuff("Might of Ursoc", true)
			and c.WearingSet(2, "GuardianT14")
	end,
	Enabled = function()
		return c.WearingSet(2, "GuardianT14")
			and not c.HasGlyph("Might of Ursoc")
	end,
	ShouldHold = function()
		if c.HasGlyph("Might of Ursoc") then
			return s.HealthPercent("player") > 70
		else
			return s.HealthPercent("player") > 50
		end
	end,
})

c.AddOptionalSpell("Might of Ursoc", "above 2 mins", {
	NoGCD = true,
	Enabled = function()
		return not c.WearingSet(2, "GuardianT14") 
			or c.HasGlyph("Might of Ursoc")
	end,
	ShouldHold = function()
		if c.HasGlyph("Might of Ursoc") then
			return s.HealthPercent("player") > 70
		else
			return s.HealthPercent("player") > 50
		end
	end,
})

c.AddOptionalSpell("Berserk", "for Guardian", {
	NoGCD = true,
	CheckFirst = function()
		return not c.AoE 
			and c.GetDebuffDuration(c.WEAKENED_BLOWS_DEBUFFS) > 13.5
	end,
	ShouldHold = function()
		return c.GetCooldown("Mangle(Bear Form)") < .5
	end,
})

c.AddOptionalSpell("Incarnation: Son of Ursoc", nil, {
	CheckFirst = function()
		return not c.AoE
	end,
	ShouldHold = function()
		return c.GetCooldown("Mangle(Bear Form)") < .5
	end,
})

c.AddOptionalSpell("Frenzied Regeneration", nil, {
	NoGCD = true,
	CheckFirst = function()
		if c.HasBuff("Savage Defense", true) and a.Rage < 90 then
			return false
		end
		
		if c.HasGlyph("Frenzied Regeneration") then
			return a.Rage >= 60 
				and s.HealthPercent("player") < 85 
				and c.IsTanking()
				and not c.IsSolo()
		else
			local maxHeal = math.max(
					2 * UnitAttackPower("player") - 4 * UnitStat("player", 2),
					2.5 * UnitStat("player", 3))
				* (1 + .1 * c.GetBuffStack("Improved Regeneration"))
			if c.WearingSet(4, "GuardianT14") then
				maxHeal = maxHeal * 1.1
			end
			return a.Rage > 0 
				and s.MaxHealth("player") - c.GetHealth("player") 
					> math.min(1, a.Rage / 60) * maxHeal
		end
	end,
})

c.AddOptionalSpell("Savage Defense", nil, {
	NoGCD = true,
	Melee = true,
	CheckFirst = function()
		return not c.HasBuff("Frenzied Regeneration") 
			and c.IsTanking()
			and not damageMode()
	end
})

c.AddOptionalSpell("Maul", "for Guardian", {
	NoGCD = true,
	CheckFirst = function()
		return (a.EmptyRage < 10 
				and (c.HasBuff("Tooth and Claw", true) or not c.IsTanking()))
			or damageMode()
	end,
})

c.AddTaunt("Growl", nil, {
	NoGCD = true,
})

c.AddSpell("Thrash(Bear Form)", nil, {
	FlashID = { "Thrash", "Thrash(Bear Form)", "Thrash(Cat Form)" },
	Melee = true,
	Cooldown = 6,
})

c.AddSpell("Thrash(Bear Form)", "for Weakened Blows", {
	FlashID = { "Thrash", "Thrash(Bear Form)", "Thrash(Cat Form)" },
	Melee = true,
	Debuff = c.WEAKENED_BLOWS_DEBUFFS,
	EarlyRefresh = 1.5,
	Cooldown = 6,
})

c.AddSpell("Thrash(Bear Form)", "for Bleed", {
	FlashID = { "Thrash", "Thrash(Bear Form)", "Thrash(Cat Form)" },
	Melee = true,
	MyDebuff = "Thrash(Bear Form)",
	UseBuffID = true,
	EarlyRefresh = 2,
	Cooldown = 6,
})

c.AddSpell("Thrash(Bear Form)", "for AoE", {
	FlashID = { "Thrash", "Thrash(Bear Form)", "Thrash(Cat Form)" },
	Melee = true,
	Cooldown = 6,
	CheckFirst = function()
		return c.AoE
	end
})

c.AddOptionalSpell("Healing Touch", "for Guardian", {
	CheckFirst = function()
		return c.HasBuff("Nature's Swiftness", false, true) 
			and c.GetHealthPercent("player") < 80
	end,
})

c.AddOptionalSpell("Rejuvenation", "for Guardian", {
	NoRangeCheck = true,
	GetDelay = function()
		return c.HasBuff("Heart of the Wild") 
			and c.GetMyBuffDuration("Rejuvenation"), 0
	end,
})

c.AddOptionalSpell("Rejuvenation", "Refresh for Guardian", {
	EarlyRefresh = 99,
	NoRangeCheck = true,
	GetDelay = function(z)
		return c.HasBuff("Heart of the Wild") 
			and c.GetMyBuffDuration("Rejuvenation") - z.EarlyRefresh, 0
	end,
})
c.ManageDotRefresh("Rejuvenation Refresh for Guardian", 3, "Rejuvenation")

c.AddSpell("Mangle(Bear Form)", "for Guardian", {
	RunFirst = function(z)
		if c.HasBuff("Berserk") or c.HasBuff("Incarnation: Son of Ursoc") then
			z.Cooldown = 0
		else
			z.Cooldown = 6
		end
	end,
})

c.AddSpell("Mangle(Bear Form)", "Delay", {
	IsMinDelayDefinition = true,
	GetDelay = function()
		return c.GetCooldown("Mangle(Bear Form)", false, 6), .5
	end
})

c.AddSpell("Faerie Fire", "for Debuff for Guardian", {
	Debuff = c.ARMOR_DEBUFFS,
	RunFirst = setFaerieID,
	SpecialGCD = "hasted",
	Cooldown = 6,
})

c.AddSpell("Enrage", nil, {
	Cooldown = 60,
	CheckFirst = function()
		return a.EmptyRage > 20 and c.IsTanking()
	end,
})

c.AddSpell("Lacerate", "for Guardian", {
	Cooldown = 3,
})

c.AddSpell("Swipe(Bear Form)", "for AoE", {
	FlashID = { "Swipe", "Swipe(Bear Form)", "Swipe(Cat Form)" },
	Melee = true,
	Cooldown = 3,
	CheckFirst = function()
		return c.AoE
	end
})

c.AddSpell("Swipe(Bear Form)", "for AoE Delay", {
	IsMinDelayDefinition = true,
	FlashID = { "Swipe", "Swipe(Bear Form)", "Swipe(Cat Form)" },
	Melee = true,
	GetDelay = function()
		return c.AoE and c.GetCooldown("Swipe(Bear Form)", false, 3), .5
	end
})

------------------------------------------------------------------- Restoration
c.RegisterForFullChannels("Tranquility")

c.AddSpell("Lifebloom", nil, {
	NoRangeCheck = true,
	EarlyRefresh = 1,
	CheckFirst = function(z)
--		if c.HasBuff("Incarnation: Tree of Life", false, false, true) then
--			c.MakeMini(z, false)
--			return true
--		end
		
		local dur = s.MyBuffDuration(z.ID, a.LifebloomTarget)
		if c.IsCastingAt(
				a.LifebloomTarget, "Regrowth", "Healing Touch", "Nourish")
			and c.GetBusyTime(true) < dur then
				
			return false
		end
		
		dur = dur - c.GetBusyTime()
		local rgCast = c.GetCastTime("Regrowth")
		local htCast = c.GetCastTime("Healing Touch")
		if dur < rgCast then
			z.FlashColor = "red"
			c.MakeMini(z, dur > z.EarlyRefresh)
			return true
		elseif dur < rgCast + z.EarlyRefresh then
			z.FlashColor = "yellow"
			c.MakeMini(z, false)
			return true
		elseif dur < htCast then
			return false
		elseif dur < htCast + z.EarlyRefresh then
			z.FlashColor = "yellow"
			c.MakeMini(z, true)
			return true
		end
	end,
})
c.ManageDotRefresh("Lifebloom", 1)

local function harmonyNeeded()
	return not c.HasBuff("Harmony")
		and not c.IsCasting("Swiftmend", "Healing Touch", "Regrowth", "Nourish")
end

c.AddOptionalSpell("Swiftmend", nil, {
	NoRangeCheck = true,
	CheckFirst = harmonyNeeded,
})

c.AddOptionalSpell("Nourish", nil, {
	NoRangeCheck = true,
	CheckFirst = harmonyNeeded,
})

local function clearcastingUp()
	return c.HasBuff("Clearcasting") 
		and not c.IsCasting("Healing Touch", "Regrowth")
end

c.AddOptionalSpell("Healing Touch", "for Restoration", {
	NoRangeCheck = true,
	CheckFirst = function()
		return clearcastingUp() or harmonyNeeded()
	end,
})

c.AddOptionalSpell("Regrowth", nil, {
	NoRangeCheck = true,
	CheckFirst = function()
		return clearcastingUp() or harmonyNeeded()
	end,
})

c.AddOptionalSpell("Innervate", nil, {
	NoRangeCheck = true,
	CheckFirst = function()
		return s.PowerPercent("player") < 80
	end
})

c.AddOptionalSpell("Cenarion Ward", "for Restoration", {
	NoRangeCheck = true,
})

